/**
 * @jest-environment jsdom
 */

import { snsProjectsCommittedStore } from "$lib/derived/sns/sns-projects.derived";
import Launchpad from "$lib/pages/Launchpad.svelte";
import { loadSnsSwapCommitments } from "$lib/services/sns.services";
import { authStore } from "$lib/stores/auth.store";
import { render, waitFor } from "@testing-library/svelte";
import {
  authStoreMock,
  mockIdentity,
  mutableMockAuthStoreSubscribe,
} from "../../mocks/auth.store.mock";
import en from "../../mocks/i18n.mock";
import {
  mockProjectSubscribe,
  mockSnsFullProject,
} from "../../mocks/sns-projects.mock";

jest.mock("$lib/services/$public/sns.services", () => {
  return {
    loadProposalsSnsCF: jest.fn().mockResolvedValue(Promise.resolve()),
  };
});

jest.mock("$lib/services/sns.services", () => {
  return {
    loadSnsSwapCommitments: jest.fn().mockResolvedValue(Promise.resolve()),
  };
});

describe("Launchpad", () => {
  describe("signed in", () => {
    jest
      .spyOn(authStore, "subscribe")
      .mockImplementation(mutableMockAuthStoreSubscribe);

    beforeAll(() =>
      authStoreMock.next({
        identity: mockIdentity,
      })
    );

    afterEach(() => jest.clearAllMocks());

    it("should render titles", () => {
      jest
        .spyOn(snsProjectsCommittedStore, "subscribe")
        .mockImplementation(mockProjectSubscribe([mockSnsFullProject]));
      const { getByText } = render(Launchpad);

      // TBU
      expect(getByText(en.sns_launchpad.open_projects)).toBeInTheDocument();
      expect(
        getByText(en.sns_launchpad.committed_projects)
      ).toBeInTheDocument();
      expect(getByText(en.sns_launchpad.proposals)).toBeInTheDocument();
    });

    it("should call loadSnsSwapCommitments", async () => {
      render(Launchpad);

      await waitFor(() => expect(loadSnsSwapCommitments).toHaveBeenCalled());
    });

    it("should not render committed project title if no committed projects", () => {
      jest
        .spyOn(snsProjectsCommittedStore, "subscribe")
        .mockImplementation(mockProjectSubscribe([]));

      const { queryByText } = render(Launchpad);

      // TBU
      expect(queryByText(en.sns_launchpad.open_projects)).toBeInTheDocument();
      expect(
        queryByText(en.sns_launchpad.committed_projects)
      ).not.toBeInTheDocument();
      expect(queryByText(en.sns_launchpad.proposals)).toBeInTheDocument();
    });
  });

  describe("not logged in", () => {
    jest
      .spyOn(authStore, "subscribe")
      .mockImplementation(mutableMockAuthStoreSubscribe);

    beforeAll(() =>
      authStoreMock.next({
        identity: undefined,
      })
    );
    it("should not call loadSnsSwapCommitments", async () => {
      render(Launchpad);

      await waitFor(() =>
        expect(loadSnsSwapCommitments).not.toHaveBeenCalled()
      );
    });
  });
});
